import itertools
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
from joblib import Parallel, delayed


class AbstractModel(ABC):
    def __init__(self, embed_dim: int, timesteps: int, max_control_cost: float, **kwargs):
        """
        Base class for all models. Your models should subclass this class.

        :param embed_dim: embedded dimension of the dynamics
        :param timesteps: timesteps per training trajectory (per action horizon)
        :param max_control_cost: maximum control cost per trajectory (per action horizon)
        :param kwargs:
        """
        self._embed_dim = embed_dim
        # NOTE: this is the timesteps of the training data; NOT the predicted trajectories
        self._timesteps = timesteps
        self._max_control_cost = max_control_cost

    @abstractmethod
    def fit(self, x: np.ndarray, **kwargs) -> None:
        """
        Trains the model. Your models must implement this method.

        :param x: (n, timesteps, embed_dim) trajectories tensor
        :param kwargs:
        :return: None
        """
        raise NotImplementedError

    def act(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """
        Determines the control for each action horizon. Your models should override this method if they use
        non-trivial control.

        :param x: (n, timesteps, embed_dim) trajectories tensor
        :param kwargs:
        :return: (n, timesteps, embed_dim) controls tensor
        """
        return np.zeros_like(x)

    def act_wrapper(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """
        Wrapper for act() called in evaluate(). Verifies control tensor is the right shape. You should NOT override this.

        :param x: (n, timesteps, embed_dim) trajectories tensor
        :param kwargs:
        :return: (n, timesteps, embed_dim) controls tensor
        """
        control = self.act(x, **kwargs)
        assert control.shape == x.shape
        return control

    @abstractmethod
    def predict(self, x0: np.ndarray, timesteps: int, **kwargs) -> np.ndarray:
        """
        Predict how initial conditions matrix evolves over a given number of timesteps. Your models must implement this method.

        NOTE: The timesteps argument can differ from the ._timesteps attribute. This allows model to train on data
        that is shorter/longer than trajectories in the test set.

        NOTE: The first coordinate of each trajectory should match the initial condition.

        :param x0: (n, embed_dim) initial conditions matrix
        :param timesteps: timesteps per predicted trajectory
        :param kwargs:
        :return: (n, timesteps, embed_dim) trajectories tensor
        """
        raise NotImplementedError

    def predict_wrapper(self, x0: np.ndarray, timesteps, **kwargs) -> np.ndarray:
        """
        Wrapper for predict() called in evaluate(). Verifies predicted trajectories tensor has the right shape.
        You should NOT override this.

        NOTE: Does not enforce that the first coordinate of each trajectory is the same as the initial condition. This
        allows DynaDojo to handle models that completely mispredict trajectory evolution.

        :param x0: (n, embed_dim) initial conditions matrix
        :param timesteps: timesteps per predicted trajectory
        :param kwargs:
        :return: (n, timesteps, embed_dim) trajectories tensor
        """
        pred = self.predict(x0, timesteps, **kwargs)
        n = x0.shape[0]
        assert pred.shape == (n, timesteps, self._embed_dim)
        return pred


class AbstractSystem(ABC):
    def __init__(self, latent_dim, embed_dim):
        """
        Base class for all systems. Your systems should subclass this class.

        NOTE: The reason we use properties and setter methods for ._latent_dim and ._embed_dim is to allow
        systems to maintain information through parameter shifts. See LDSSystem in './systems/lds.py' for a principled
        usage example of the setter methods.
        """
        self._latent_dim = latent_dim
        self._embed_dim = embed_dim

    @property
    def latent_dim(self):
        return self._latent_dim

    @property
    def embed_dim(self):
        return self._embed_dim

    @latent_dim.setter
    def latent_dim(self, value):
        self._latent_dim = value

    @embed_dim.setter
    def embed_dim(self, value):
        self._embed_dim = value

    @abstractmethod
    def make_init_conds(self, n: int, in_dist=True) -> np.ndarray:
        """
        Abstract method to generate initial conditions. Your system should override this method.

        NOTE: Systems developers determine what counts as in vs out-of-distribution. DynaDojo doesn't provide
        any verification that this distinction makes sense or even exists.

        :param n: number of initial conditions
        :param in_dist: Boolean. If True, generate in-distribution initial conditions. Defaults to True. If False,
        generate out-of-distribution initial conditions.
        :return: (n, embed_dim) initial conditions matrix
        """
        raise NotImplementedError

    def make_init_conds_wrapper(self, n: int, in_dist=True):
        """
        Wrapper for make_init_conds() called in evaluate(). Verifies initial condition matrix is the right shape.
        You should NOT override this.

        :param n: number of initial conditions
        :param in_dist: Boolean. If True, generate in-distribution initial conditions. Defaults to True. If False,
        generate out-of-distribution initial conditions.
        :return: (n, embed_dim) initial conditions matrix
        """
        init_conds = self.make_init_conds(n, in_dist)
        assert init_conds.shape == (n, self.embed_dim)
        return init_conds

    @abstractmethod
    def make_data(self, init_conds: np.ndarray, control: np.ndarray, timesteps: int, noisy=False) -> np.ndarray:
        """
        Abstract method that makes trajectories from initial conditions. Your system must override this method.

        :param init_conds: (n, embed_dim) initial conditions matrix
        :param control: (n, timesteps, embed_dim) controls tensor
        :param timesteps: timesteps per training trajectory (per action horizon)
        :param noisy: Boolean. If True, add noise to trajectories. Defaults to False. If False, no noise is added.
        :return: (n, timesteps, embed_dim) trajectories tensor
        """
        raise NotImplementedError

    def make_data_wrapper(self, init_conds: np.ndarray, control: np.ndarray = None, timesteps: int = 1,
                          noisy=False) -> np.ndarray:
        """
        Wraps make_data(). Checks that trajectories tensor has the proper shape. You should NOT override this.

        :param init_conds: (n, embed_dim) initial conditions matrix
        :param control: (n, timesteps, embed_dim) controls tensor
        :param timesteps: timesteps per training trajectory (per action horizon)
        :param noisy: Boolean. If True, add noise to trajectories. Defaults to False. If False, no noise is added.
        :return: (n, timesteps, embed_dim) trajectories tensor
        """
        assert timesteps > 0
        assert init_conds.ndim == 2 and init_conds.shape[1] == self.embed_dim
        n = init_conds.shape[0]
        if control is None:
            control = np.zeros((n, timesteps, self.embed_dim))
        assert control.shape == (n, timesteps, self.embed_dim)
        data = self.make_data(init_conds=init_conds,
                              control=control, timesteps=timesteps, noisy=noisy)
        assert data.shape == (n, timesteps, self.embed_dim)
        return data

    @abstractmethod
    def calc_error(self, x, y) -> float:
        """
        Calculates the error between two tensors of trajectories. Your systems must implement this.

        :param x: (n, timesteps, embed_dim) trajectories tensor
        :param y: (n, timesteps, embed_dim) trajectories tensor
        :return: Float. The error between x and y.
        """
        raise NotImplementedError

    def calc_error_wrapper(self, x, y) -> float:
        """
        Wraps calc_error. Checks that calc_error is called with properly-shaped x and y.
        Your systems should NOT override this.

        :param x: (n, timesteps, embed_dim) trajectories tensor
        :param y: (n, timesteps, embed_dim) trajectories tensor
        :return: Float. The error between x and y.
        """
        assert x.shape == y.shape
        return self.calc_error(x, y)

    @abstractmethod
    def calc_control_cost(self, control: np.ndarray) -> np.ndarray:
        """
        Calculate the control cost for each control TRAJECTORY (i.e., calculates the costs for every
        control matrix, not for the whole tensor). Your systems must implement this.

        :param control: (n, timesteps, embed_dim) controls tensor
        :return: (n,) control costs vector
        """
        raise NotImplementedError

    def calc_control_cost_wrapper(self, control: np.ndarray) -> np.ndarray:
        """
        Wraps calc_control_cost(). Your systems should NOT override this.

        :param control: (n, timesteps, embed_dim) controls tensor
        :return: (n,) control costs vector
        """
        assert control.shape[2] == self.embed_dim and control.ndim == 3
        cost = self.calc_control_cost(control)
        assert cost.shape == (len(control),)
        return cost


class Task:
    def __init__(self,
                 N: list[int],
                 L: list[int],
                 E: list[int] | int | None,
                 T: list[int],
                 max_control_cost_per_dim: int,
                 control_horizons: int,
                 system_cls: type[AbstractSystem],
                 reps: int,
                 test_examples: int,
                 test_timesteps: int,
                 system_kwargs: dict = None,
                 ):
        """


        :param N: train sizes
        :param L: latent dimensions
        :param E: embedded dimensions. Optional. If list, then evaluate iterates across embedded dimensions.
        If int, then evaluate uses a fixed embedded dimension. If None, then evaluate sets the embedded dimension
        equal to the latent dimension.
        :param T: timesteps
        :param max_control_cost_per_dim: max control cost per control trajectory
        :param control_horizons: number of times to generate training data with control
        :param system_cls: class constructor (NOT instance) for a concrete system
        :param reps: number of times to repeat each experiment
        :param test_examples: test size
        :param test_timesteps: test timesteps
        :param system_kwargs:
        """
        assert control_horizons >= 0

        self._id = itertools.count()
        self._N = N
        self._L = L
        self._E = E
        self._T = T
        self._max_control_cost_per_dim = max_control_cost_per_dim
        self._system_cls = system_cls
        self._system_kwargs = system_kwargs or {}
        self._control_horizons = control_horizons
        self._reps = reps
        self._test_examples = test_examples
        self._test_timesteps = test_timesteps

    def evaluate(self,
                 model_cls: type[AbstractModel],
                 model_kwargs: dict = None,
                 fit_kwargs: dict = None,
                 act_kwargs: dict = None,
                 in_dist=True,
                 noisy=False,
                 id=None,
                 ) -> pd.DataFrame:
        """
        Evaluates a model class (NOT an instance) on a dynamical system over a set of experimental parameters.

        :param model_cls: model class to be evaluated
        :param model_kwargs:
        :param fit_kwargs:
        :param act_kwargs:
        :param in_dist: Boolean. If True, generate in-distribution initial conditions for the test set. Defaults to True.
        If False, generate out-of-distribution initial conditions for the test set.
        :param noisy: Boolean. If True, add noise to train set. Defaults to False. If False, no noise is added.
        :param id: model ID associated with evaluation results in returned DataFrame
        :return: a pandas DataFrame with experimental results
        """

        model_kwargs = model_kwargs or {}
        fit_kwargs = fit_kwargs or {}
        act_kwargs = act_kwargs or {}

        def do_rep1(rep_id):
            """
            Handles case when E is an array
            """
            result = {k: [] for k in ["rep", "n", "latent_dim",
                                      "embed_dim", "timesteps", "control_horizons", "loss", "total_cost"]}
            system = None
            for n, d, timesteps in itertools.product(self._N, zip(self._L, self._E), self._T):
                latent_dim, embed_dim = d
                if embed_dim < latent_dim:
                    return
                system = self._set_system(system, latent_dim, embed_dim)
                self._do_rep(rep_id, id, result, system, n, latent_dim, embed_dim, timesteps, model_cls, model_kwargs, fit_kwargs,
                             act_kwargs, in_dist, noisy)
            return pd.DataFrame(result)

        def do_rep2(rep_id):
            """
            Handles case when E is a None
            """
            result = {k: [] for k in ["rep", "n", "latent_dim",
                                      "embed_dim", "timesteps", "control_horizons", "loss", "total_cost"]}
            system = None
            for n, latent_dim, timesteps in itertools.product(self._N, self._L, self._T):
                embed_dim = latent_dim
                system = self._set_system(system, latent_dim, embed_dim)
                self._do_rep(rep_id, id, result, system, n, latent_dim, embed_dim, timesteps, model_cls, model_kwargs, fit_kwargs,
                             act_kwargs, in_dist, noisy)
            return pd.DataFrame(result)

        def do_rep3(rep_id):
            """
            Handles case when E is a constant
            """
            result = {k: [] for k in ["rep", "n", "latent_dim",
                                      "embed_dim", "timesteps", "control_horizons", "loss", "total_cost"]}
            system = None
            for n, latent_dim, timesteps in itertools.product(self._N, self._L, self._T):
                embed_dim = self._E
                if embed_dim < latent_dim:
                    return
                system = self._set_system(system, latent_dim, embed_dim)
                self._do_rep(rep_id, id, result, system, n, latent_dim, embed_dim, timesteps, model_cls, model_kwargs, fit_kwargs,
                             act_kwargs, in_dist, noisy)
            return pd.DataFrame(result)

        if isinstance(self._E, list):
            do_rep = do_rep1
        elif self._E is None:
            do_rep = do_rep2
        elif isinstance(self._E, int):
            do_rep = do_rep3
        else:
            raise TypeError("E must of type List[int], int, or None.")

        data = Parallel(n_jobs=4, timeout=1e6)(delayed(do_rep)(rep_id)
                                               for rep_id in range(self._reps))
        data = pd.concat(data)
        data["id"] = id or next(self._id)
        return data

    def _set_system(self, system, latent_dim, embed_dim):
        if system is None:
            system = self._system_cls(
                latent_dim, embed_dim, **self._system_kwargs)
        if latent_dim != system.latent_dim:
            system.latent_dim = latent_dim
        if embed_dim != system.embed_dim:
            system.embed_dim = embed_dim
        return system

    def _gen_trainset(self, system, n: int, timesteps: int, noisy=False):
        train_init_conds = system.make_init_conds_wrapper(n)
        return system.make_data_wrapper(train_init_conds, timesteps=timesteps, noisy=noisy)

    def _gen_testset(self, system, in_dist=True):
        test_init_conds = system.make_init_conds_wrapper(self._test_examples, in_dist)
        return system.make_data_wrapper(test_init_conds, timesteps=self._test_timesteps)

    def _fit_model(self, system, model, x: np.ndarray, timesteps: int,  max_control_cost: int,  fit_kwargs: dict = None,
                   act_kwargs: dict = None, noisy=False) -> int:
        total_cost = 0
        model.fit(x, **fit_kwargs)

        for _ in range(self._control_horizons):
            control = model.act_wrapper(x, **act_kwargs)
            cost = system.calc_control_cost_wrapper(control)
            total_cost += cost
            assert np.all(cost <= max_control_cost), "Control cost exceeded!"
            x = system.make_data_wrapper(
                init_conds=x[:, -1], control=control, timesteps=timesteps, noisy=noisy)
            model.fit(x, **fit_kwargs)

        return total_cost

    def _append_result(self, result, rep_id, n, latent_dim, embed_dim, timesteps, loss, total_cost):
        result['rep'].append(rep_id)
        result['n'].append(n)
        result['latent_dim'].append(latent_dim)
        result['embed_dim'].append(embed_dim)
        result['timesteps'].append(timesteps)
        result['control_horizons'].append(self._control_horizons)
        result['loss'].append(loss)
        result['total_cost'].append(total_cost)

    def _do_rep(self,
                rep_id: int,
                id: str | int,
                result: dict,
                system: AbstractSystem,
                n: int,
                latent_dim: int,
                embed_dim: int,
                timesteps: int,
                model_cls: type[AbstractModel],
                model_kwargs: dict = None,
                fit_kwargs: dict = None,
                act_kwargs: dict = None,
                in_dist=True,
                noisy=False,
                ):
        max_control_cost = self._max_control_cost_per_dim * latent_dim
        print(f"{n=}, {latent_dim=}, {embed_dim=}, {timesteps=}, control_horizons={self._control_horizons}, { rep_id=}, {id=}")

        # Create model and data
        model = model_cls(embed_dim, timesteps, max_control_cost, **model_kwargs)
        x = self._gen_trainset(system, n, timesteps, noisy)
        total_cost = self._fit_model(system, model, x, timesteps, max_control_cost, fit_kwargs, act_kwargs, noisy)
        test = self._gen_testset(system, in_dist)
        pred = model.predict_wrapper(test[:, 0], self._test_timesteps)
        loss = system.calc_error_wrapper(pred, test)
        self._append_result(result, rep_id, n, latent_dim, embed_dim, timesteps, loss, total_cost)



import numpy as np
import torch
import torch.nn as nn
from torchdiffeq import odeint

from ..abstractions import AbstractModel


class ODE(AbstractModel):
    def __init__(self, embed_dim: int, timesteps: int, max_control_cost: float, lr=3e-2, **kwargs):
        super().__init__(embed_dim, timesteps, max_control_cost, **kwargs)

        self.model = nn.Sequential(nn.Linear(embed_dim, 32), nn.Softplus(), nn.Linear(32, embed_dim))
        self.lr = lr
        self.mse_loss = nn.MSELoss()
        self.opt = torch.optim.Adam(self.model.parameters(), self.lr)
    
    def forward(self, t, state):
        dx = self.model(state)
        return dx

    def predict(self, x0: np.ndarray, timesteps: int, **kwargs) -> np.ndarray:
        pass

    def fit(self, x: np.ndarray, epochs=100, **kwargs):
        x = torch.tensor(x, dtype=torch.float32)
        state = x[:, 0, :]
        t = torch.linspace(0.0, self._timesteps, self._timesteps)
        for _ in range(epochs):
            self.opt.zero_grad()
            pred = odeint(self.forward, state, t, method='rk4')
            pred = pred.transpose(0, 1)
            loss = self.mse_loss(pred, x).float()
            loss.backward()
            self.opt.step()
    
    def predict(self, x0: np.ndarray, timesteps: int, **kwargs) -> np.ndarray:
        x0 = torch.tensor(x0, dtype=torch.float32)
        t = torch.linspace(0.0, timesteps, timesteps)
        return odeint(self, x0, t).detach().numpy()



import numpy as np
import scipy as sp
from scipy.integrate import solve_ivp
from scipy.stats import ortho_group

from ..abstractions import AbstractSystem

RNG = np.random.default_rng()


class LDSSystem(AbstractSystem):
    def __init__(self, latent_dim, embed_dim,
                 # negative eigenvalues produce stable linear system (https://en.wikipedia.org/wiki/Stability_theory)
                 A_eigval_range=(-5, 0),
                 A_eigvec_range=(-1, 1),
                 BC_sv_range=(-1, 1),
                 init_cond_in_dist_range=(0, 10),
                 init_cond_out_dist_range=(-10, 0),
                 scale=0.01
                 ):
        super().__init__(latent_dim, embed_dim)
        self._A_eigval_range = A_eigval_range
        self._A_eigvec_range = A_eigvec_range
        self._BC_sv_range = BC_sv_range
        self._init_cond_in_dist_range = init_cond_in_dist_range
        self._init_cond_out_dist_range = init_cond_out_dist_range
        self._scale = scale

        self.A = self._make_A()
        self.B = self._make_BC()
        self.C = self._make_BC()

    @staticmethod
    def _eigenvalues_to_matrix(dim, eig_range, eigvec_range):
        # use eigendecomposition to produce a real-valued matrix with target eigenvalues
        if RNG.uniform() <= 0:  # TODO: change
            # 50% chance to have all real eigenvalues
            eigenvalues = RNG.uniform(*eig_range, size=dim)
            Q = RNG.uniform(*eigvec_range, size=(dim, dim))
            Q = sp.linalg.orth(Q)  # orthogonal matrices preserve operator norms
        else:
            # 50% chance to have some complex eigenvalues
            num_complex = RNG.integers(0, dim, endpoint=True)
            if num_complex % 2 == 1:
                num_complex -= 1
            num_real = dim - num_complex

            real_eigenvalues = RNG.uniform(*eig_range, size=num_real)
            real_eigenvectors = RNG.uniform(*eigvec_range, size=(dim, num_real))

            # complex eigenvectors and eigenvalues must have conjugate pairs in the diagonalization for A to be real
            complex_eigenvalues = RNG.uniform(*eig_range, size=num_complex // 2) + 1j * RNG.uniform(*eig_range, size=num_complex // 2)
            complex_eigenvectors = RNG.uniform(*eigvec_range, size=(dim, num_complex // 2)) + 1j * RNG.uniform(*eigvec_range, size=(dim, num_complex // 2))

            eigenvalues = np.concatenate((real_eigenvalues, complex_eigenvalues, complex_eigenvalues.conjugate()))
            Q = np.hstack((real_eigenvectors, complex_eigenvectors, complex_eigenvectors.conjugate()))

            # shuffle columns
            pi = RNG.permutation(dim)
            Q = Q[pi]
            eigenvalues = eigenvalues[pi]

        M = Q @ np.diag(eigenvalues) @ np.linalg.inv(Q)
        return M.real

    @staticmethod
    def _singular_values_to_matrix(m, n, sv_range):
        U = ortho_group.rvs(m)
        sigma = np.eye(m, n) * RNG.uniform(*sv_range, size=n)
        V = ortho_group.rvs(n)
        M = U @ sigma @ V
        return M

    def _make_A(self):
        return self._eigenvalues_to_matrix(self.latent_dim, self._A_eigval_range, self._A_eigvec_range)

    def _make_BC(self):
        return self._singular_values_to_matrix(self.latent_dim, self.embed_dim, self._BC_sv_range)

    @AbstractSystem.embed_dim.setter
    def embed_dim(self, value):
        self._embed_dim = value
        self.B = self._make_BC()
        self.C = self._make_BC()

    @AbstractSystem.latent_dim.setter
    def latent_dim(self, value):
        self._latent_dim = value
        self.A = self._make_A()
        self.B = self._make_BC()
        self.C = self._make_BC()

    def make_init_conds(self, n: int, in_dist=True) -> np.ndarray:
        init_cond_range = self._init_cond_in_dist_range if in_dist else self._init_cond_out_dist_range
        return RNG.uniform(*init_cond_range, (n, self.embed_dim))

    def make_data(self, init_conds: np.ndarray, control: np.ndarray, timesteps: int, noisy=False) -> np.ndarray:
        data = []
        init_conds = init_conds @ np.linalg.pinv(self.C)
        time = np.linspace(0, 1, num=timesteps)

        def dynamics(t, x, u):
            i = np.argmin(np.abs(t - time))
            if noisy:
                return self.A @ x + self.B @ u[i] + RNG.normal(scale=self._scale, size=(self.latent_dim))
            else:
                return self.A @ x + self.B @ u[i]

        for x0, u in zip(init_conds, control):
            sol = solve_ivp(dynamics, t_span=[0, 1], y0=x0, t_eval=time, dense_output=True, args=(u,))
            data.append(sol.y)
        data = np.transpose(np.array(data), axes=(0, 2, 1)) @ self.C
        return data

    def calc_error(self, x, y) -> float:
        # TODO: add more details
        error = x - y
        return np.mean(error ** 2) / self.embed_dim

    def calc_control_cost(self, control: np.ndarray) -> float:
        return np.linalg.norm(control, axis=(1, 2), ord=2) / self.embed_dim  # TODO: ask Max



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

def _assign_labels(data, labels):
    assert len(data) == len(labels)
    for frameidx, frame in enumerate(data):
        data[frameidx] = frame.assign(id=labels[frameidx])


def plot_metric(data, xcol, ycol, idlabels=None, xlabel=None, ylabel=None, hue="id", log=True, estimator=np.median, errorbar=("pi", 50)):
    if idlabels:
        _assign_labels(data, idlabels)

    sns.set_context("paper")
    sns.set_theme(style="ticks")
    sns.despine()

    if not isinstance(data, pd.DataFrame):
        data = pd.concat(data)

   
    ax = sns.catplot(data=data, x=xcol, y=ycol, hue=hue, estimator=estimator, kind="point", errorbar=errorbar)

    ax.set(xticks=range(len(data[xcol].unique())))
    if log:
        ax.set(yscale="log")

    if xlabel:
       ax.set(xlabel=xlabel)
    if ylabel:
       ax.set(ylabel=ylabel)

    plt.show()


def plot_target_loss(data, xcol, ycol, idlabels=None, xlabel=None, ylabel=None, hue="id", log=True, estimator=np.median, errorbar=("pi", 50), error_col="loss", target_loss=0.1):
    if idlabels:
        _assign_labels(data, idlabels)

    if not isinstance(data, pd.DataFrame):
        data = pd.concat(data)

    filtered = data[data[error_col] <= target_loss]
    # for each rep in the x dim, get the lowest y that was successful
    successes = filtered.loc[filtered.groupby(["id", "rep", xcol])[ycol].idxmin()].reset_index(drop=True)

    assert (len(successes) > 0)
    plot_metric(successes, xcol, ycol, None, xlabel, ylabel, hue, log, estimator, errorbar)



import numpy as np
import torch
import torch.nn as nn

from ..abstractions import AbstractModel


class CNN(AbstractModel):
    def __init__(self, embed_dim: int, timesteps: int, max_control_cost: float, lr=6e-2, k=3, **kwargs):
        super().__init__(embed_dim, timesteps, max_control_cost, **kwargs)

        self.lr = lr
        self.k = k
        self.c1 = nn.Sequential(
            nn.Conv1d(1, 32, self.k, padding='same'),
            nn.MaxPool1d(self.k),
            nn.BatchNorm1d(32),
            nn.Conv1d(32, 1, self.k, padding='same'),
            nn.BatchNorm1d(1)
        )
        n_channels = self.c1(torch.empty(1, 1, self._embed_dim)).size(-1)
        self.lin = nn.Sequential(nn.Linear(n_channels, embed_dim), nn.Sigmoid())

    def fit(self, x: np.ndarray, epochs=100, **kwargs):
        state = torch.tensor(x, dtype=torch.float32)
        batch_size = len(state)

        opt = torch.optim.Adam(self.c1.parameters(), self.lr)
        loss_BCEL = nn.BCELoss()
        lossMSE = nn.MSELoss()

        for i in range(epochs):
            opt.zero_grad()
            pred_states = state[:, 0, :].unsqueeze(1)
            loss = 0.0

            for t in range(self._timesteps - 1):
                state_t = state[:, t, :]
                state_t = torch.reshape(state_t, (batch_size, self._embed_dim)).unsqueeze(1)

                next_state = self.lin(self.c1(state_t))  # doesn't even call forward

                loss += lossMSE((next_state * (state[:, t, :].unsqueeze(1))), state[:, t+1, :].unsqueeze(1)) #loss_BCEL(next_state, state[:, t + 1, :].unsqueeze(1))

            # print(loss.item())
            loss.backward()
            opt.step()

    def predict(self, x0: np.ndarray, timesteps: int, **kwargs) -> np.ndarray:
        state = torch.tensor(x0, dtype=torch.float32)
        pred = state.unsqueeze(1)

        for t in range(timesteps - 1):
            state_t = pred[:, -1, :].unsqueeze(1)

            next_prob = self.lin(self.c1(state_t))  # doesn't even call forward

            next_state = torch.distributions.bernoulli.Bernoulli(next_prob).sample()
            pred = torch.cat((pred, next_state), dim=1)

        return pred.detach().numpy()


class _CNN(nn.Module): #change to inherit from model instead
    def __init__(self, latent_dim, embed_dim, timesteps, lr = 6e-2, epochs = 100):
        super().__init__()
        self._latent_dim = latent_dim
        self._embed_dim = embed_dim
        self._timesteps = timesteps
        self.lr = lr
        self.epochs = epochs
        self.k = 5
        
        self.c1 = nn.Sequential(nn.Conv1d(1, 32, self.k, padding='same'), nn.MaxPool1d(self.k), nn.Conv1d(32, 1, self.k, padding='same'))
        
        n_channels = self.c1(torch.empty(1, 1, self._embed_dim)).size(-1)
        self.lin = nn.Sequential(nn.Linear(n_channels, embed_dim), nn.Sigmoid())

    def forward(self, x):
        return self.lin(self.c1(x))
    
    def fit(self, x):
        
        state = torch.tensor(x, dtype=torch.float32) 
        batch_size = len(state)
        
        opt = torch.optim.Adam(self.c1.parameters(), self.lr)
        loss_BCEL = nn.BCELoss()

        for i in range(self.epochs):
            opt.zero_grad()
            pred_states = state[:, 0, :].unsqueeze(1)
            loss = 0.0
           
            for t in range(self._timesteps-1):
                state_t = state[:, t, :]
                state_t = torch.reshape(state_t, (batch_size, self._embed_dim)).unsqueeze(1)
                
                next_state = self.lin(self.c1(state_t)) #doesnt even call forward
              
                loss += loss_BCEL(next_state, state[:, t+1, :].unsqueeze(1))
            
            print(loss.item())
            loss.backward()
            opt.step()

    def predict(self, x0, timesteps):
        
        state = torch.tensor(x0, dtype=torch.float32) 
        pred = state.unsqueeze(1)
           
        for t in range(timesteps-1):
            state_t = pred[:, -1, :].unsqueeze(1)
            
            next_prob = self.lin(self.c1(state_t)) #doesnt even call forward

            next_state = torch.distributions.bernoulli.Bernoulli(next_prob).sample()
            pred = torch.cat((pred, next_state), dim=1)

        return pred.detach().numpy()



import numpy as np
import tensorflow as tf

from ..abstractions import AbstractModel


class DNN(AbstractModel):

    def __init__(self, embed_dim, timesteps, max_control_cost, activation=None, **kwargs):
        super().__init__(embed_dim, timesteps, max_control_cost, **kwargs)
        self.model = tf.keras.Sequential([
            tf.keras.Input(shape=(None, embed_dim)),
            tf.keras.layers.Dense(30, activation=activation, kernel_regularizer="l2"),
            tf.keras.layers.Dense(30, activation=activation, kernel_regularizer="l2"),
            tf.keras.layers.Dense(30, activation=activation, kernel_regularizer="l2"),
            tf.keras.layers.Dense(10, activation="linear"),
            tf.keras.layers.Dense(30, activation=activation, kernel_regularizer="l2"),
            tf.keras.layers.Dense(30, activation=activation, kernel_regularizer="l2"),
            tf.keras.layers.Dense(30, activation=activation, kernel_regularizer="l2"),
            tf.keras.layers.Dense(embed_dim, kernel_regularizer="l2")
        ])
        self.model.compile(optimizer="adam", loss="mean_squared_error")

    def fit(self, x: np.ndarray, epochs=20, verbose=0, **kwargs):
        head = x[:, :-1, :]
        tail = x[:, 1:, :]
        self.model.fit(head, tail, epochs=epochs, verbose=verbose)

    def predict(self, x0: np.ndarray, timesteps: int, **kwargs) -> np.ndarray:
        preds = [x0]
        for _ in range(timesteps - 1):
            preds.append(self.model.predict(preds[-1], verbose=0))
        preds = np.array(preds).squeeze().transpose(1, 0, 2)
        return preds





import cellpylib as cpl
import numpy as np
from joblib import Parallel, delayed

from ..abstractions import AbstractSystem

RNG = np.random.default_rng()


class CASystem(AbstractSystem):
    def __init__(self, latent_dim, embed_dim, in_dist_p=0.25, out_dist_p=0.75, mutation_p=0.00):
        super().__init__(latent_dim, embed_dim)
        lambda_val = RNG.uniform()
        self.rule_table, _, _ = cpl.random_rule_table(lambda_val=lambda_val, k=2, r=self.latent_dim,
                                                      strong_quiescence=True,
                                                      isotropic=True)
        self._in_dist_p = in_dist_p
        self._out_dist_p = out_dist_p
        self._mutation_p = mutation_p

    @AbstractSystem.latent_dim.setter
    def latent_dim(self, value):
        self._latent_dim = value
        lambda_val = RNG.uniform()
        self.rule_table, _, _ = cpl.random_rule_table(lambda_val=lambda_val, k=2, r=self.latent_dim,
                                                      strong_quiescence=True,
                                                      isotropic=True)

    def make_init_conds(self, n: int, in_dist=True) -> np.ndarray:
        if in_dist:
            return RNG.binomial(1, self._in_dist_p, size=(n, self.embed_dim))
        else:
            return RNG.binomial(1, self._out_dist_p, size=(n, self.embed_dim))

    def make_data(self, init_conds: np.ndarray, control: np.ndarray, timesteps: int, noisy=False) -> np.ndarray:
        data = []

       # print(f'self.rule_table: {self.rule_table}')
        def get_trajectory(x0, u):
            cellular_automata = np.clip([x0 + u[0]], 0, 1).astype(np.int32)
            for t in range(1, timesteps):
                cellular_automata = cpl.evolve(cellular_automata,
                                               timesteps=2,
                                               apply_rule=lambda n, c, t: cpl.table_rule(n, self.rule_table),
                                               r=self.latent_dim)
                cellular_automata[-1] = np.clip(cellular_automata[-1] + u[t], 0, 1).astype(np.int32)
                if noisy:
                    mask = RNG.binomial(1, self._mutation_p, size=(self.embed_dim,)).astype(bool)
                    cellular_automata[-1][mask] = (~cellular_automata[-1][mask].astype(bool)).astype(np.int32)
            return cellular_automata

        data = Parallel(n_jobs=4)(delayed(get_trajectory)(x0, u) for x0, u in zip(init_conds, control))
        data = np.array(data)
        return data

    def calc_error(self, x, y):
        # averaged across all samples and all predicted timesteps
        return 1 - (np.count_nonzero(x == y) / self.embed_dim) / len(y) / len(y[1])

    def calc_control_cost(self, control: np.ndarray) -> float:
        return np.sum(control, axis=(1, 2))


import numpy as np
import torch
import torch.nn as nn
from torchdiffeq import odeint

from ..abstractions import AbstractModel

class AugODE(AbstractModel):
    def __init__(self, embed_dim: int, timesteps: int, max_control_cost: float, lr=3e-2, aug_dim = 2, **kwargs):
        super().__init__(embed_dim, timesteps, max_control_cost)

        self.model = nn.Sequential(nn.Linear(embed_dim + aug_dim, 32), nn.Softplus(), nn.Linear(32, embed_dim))
        self.lr = lr
        self.aug_dim = aug_dim
        self.mse_loss = nn.MSELoss()
        self.opt = torch.optim.Adam(self.model.parameters(), self.lr)

    def forward(self, t, state):
        #zeros = torch.zeros(self.aug_dim)
        #aug_state = torch.cat((state, zeros)) #naive way does not generalize
        aug_state = torch.nn.functional.pad(input=state, pad=(0, self.aug_dim), mode='constant', value=0) #is this correct and does it generalize to different data shapes?
        
        dx = self.model(aug_state)
        return dx

    def fit(self, x: np.ndarray, epochs=100, **kwargs):
        x = torch.tensor(x, dtype=torch.float32)
        state = x[:, 0, :]
        step = end = epochs/self._timesteps
        
        for _ in range(epochs):
            if _ % step == 0:
                t = torch.linspace(0.0, end, self._timesteps)
                end += step
            self.opt.zero_grad()
            pred = odeint(self.forward, state, t, method='midpoint')
            pred = pred.transpose(0, 1)
            loss = self.mse_loss(pred, x).float()
            loss.backward()
            self.opt.step()

    def predict(self, x0, timesteps):
        x0 = torch.tensor(x0, dtype=torch.float32)
        t = torch.linspace(0.0, timesteps, timesteps)
        return odeint(self, x0, t).detach().numpy()



__author__ = "Logan Mondal Bhamidipaty, Tommy Anthony Bruzzese, Max Kanwal, Rami Ratl Mrad"
__copyright__ = "Copyright 2023"
__credits__ = ["Logan Mondal Bhamidipaty", "Tommy Anthony Bruzzese", "Max Kanwal", "Rami Ratl Mrad"]
__license__ = "MIT"

from . import systems
from . import utils
from . import baselines
from . import abstractions
from . import challenges



from ..abstractions import AbstractSystem
import torch
import torch.nn as nn
from torchdiffeq import odeint
import networkx as nx
import numpy as np
import random

RNG = np.random.default_rng()


class CTLNSystem(AbstractSystem):
    def __init__(self, latent_dim, embed_dim, p):
        super().__init__(latent_dim, embed_dim)
        self._latent_dim = latent_dim
        self._embed_dim = embed_dim
        self.relu = nn.ReLU()
        self._p = p
        self._nodes = latent_dim

    def make_init_conds(self, n: int, in_dist=True) -> np.ndarray:
        self._delta = random.uniform(0, 1) #what should the upper bound on the randomly generated constants be?
        self._epsilon = random.uniform(0, self._delta/(self._delta+1))
        graph = self._make_graph(self._nodes, self._p)
        x0 = torch.Tensor(RNG.uniform(0, 1, size=(n, self.embed_dim)))
        time = torch.tensor([random.uniform(0, 1) for _ in range(self._nodes)])
        b = torch.Tensor(RNG.uniform(0, 1, size=(self.embed_dim, n)))
        self._state = (x0, time, graph, b)
        return x0

    def _make_graph(self, nodes, p):
        g = nx.erdos_renyi_graph(nodes, p)
        edges = nx.adjacency_matrix(g).todense()
        edges_comp = nx.adjacency_matrix(nx.complement(g)).todense()
        graph = (-edges_comp - self._epsilon) + (-edges + self._delta)      
        np.fill_diagonal(graph, 0)
        return torch.tensor(graph)


    def make_data(self, init_conds: np.ndarray, control: np.ndarray, timesteps: int, noisy=False) -> np.ndarray:
        data = []
        time = torch.linspace(0, 1, timesteps)

        def dynamics(t, state):
            x, T, W, b = state
            y = W @ x.T + b
            dx = (-x.T + self.relu(y)) / T[:, None]
            dT = torch.zeros_like(T)
            dW = torch.zeros_like(W)
            db = torch.zeros_like(b)
            return (dx, dT, dW, db)

        sol = odeint(dynamics, self._state, time, method='rk4')
        data.append(sol[0])
        # result = sol[0].unsqueeze(0).detach().numpy()
        result = np.transpose(sol[0].detach().numpy(), axes=(1, 0, 2))
        return result

    def calc_error(self, x, y) -> float:
        error = x - y
        return np.mean(error ** 2) / self.embed_dim

    def calc_control_cost(self, control: np.ndarray) -> float:
        return np.linalg.norm(control, axis=(1, 2), ord=2)



import numpy as np
import math

from ..abstractions import AbstractModel


class LowestPossibleRadius(AbstractModel):
    def __init__(self, embed_dim, timesteps, max_control_cost):
        super().__init__(embed_dim, timesteps, max_control_cost)
        self.currRadius = 1
        self.radiiTables = {}
        self.radiiTables[self.currRadius] = self.generateRadiusTable(self.currRadius)
        self.ROW_LENGTH = embed_dim
        self._max_control_cost = max_control_cost

    def generateCombos(self, n):
        if not n:
            return

        for i in range(2**n):
            s = bin(i)[2:]
            s = "0" * (n-len(s)) + s
            yield s

    def generateRadiusTable(self, radius):
        allCombos = self.generateCombos((radius*2)+1)
        return dict.fromkeys(allCombos)

    def isValidRadius(self, radius, samples) -> bool:
        for sample in samples:
            for stepidx, step in enumerate(sample):
                if stepidx == 0:
                    continue

                for cellidx, cell in enumerate(step):
                    neighborhood = ""

                    # to the left
                    for neg_x in range(radius, 0, -1):
                        neighborhood += str(sample[stepidx - 1][cellidx - neg_x])

                    # directly above
                    neighborhood += str(sample[stepidx - 1][cellidx])

                    # to the right
                    for pos_x in range(1, radius+1):
                        pos = (cellidx + pos_x) % self.ROW_LENGTH
                        neighborhood += str(sample[stepidx - 1][pos])

                    if (self.radiiTables[radius][neighborhood]) == None:
                        self.radiiTables[radius][neighborhood] = cell
                    else:
                        if self.radiiTables[radius][neighborhood] != cell:
                            return False
        return True

    def fit(self, samples):
        while (not self.isValidRadius(self.currRadius, samples)):
            newRadius = self.currRadius+1
            self.radiiTables[newRadius] = self.generateRadiusTable(newRadius)
            self.currRadius = newRadius

    def act(self, x, **kwargs):
        control = []
        lastState = x[:, -1, :]
        control_mag = {}  # constraint is for each sample

        # find a smart control for 1st next state of traj
        for sampleidx, sample in enumerate(lastState):
            tempControl = []
            for _ in range(math.floor(self._embed_dim / ((self.currRadius*2) + 1))):
                cellidx = self.currRadius
                neighborhood = ""

                # left of cell
                for neg_x in range(self.currRadius, 0, -1):
                    neighborhood += str(sample[cellidx - neg_x])

                # the cell
                neighborhood += str(sample[cellidx])

                # right of cell
                for pos_x in range(1, self.currRadius+1):
                    pos = (cellidx + pos_x) % self.ROW_LENGTH
                    neighborhood += str(sample[pos])

                # if key is seen, replace with unseen
                if (self.radiiTables[self.currRadius][neighborhood]) != None:
                    unseenKeys = [
                        key for key, value in self.radiiTables[self.currRadius].items() if value is None]
                    if unseenKeys:
                        desiredKey = np.random.choice(unseenKeys)

                        for idx, element in enumerate(neighborhood):
                            if sampleidx not in control_mag:
                                control_mag[sampleidx] = 0
                            if control_mag[sampleidx] >= self._max_control_cost:
                                tempControl.append(0)
                            else:
                                if element == desiredKey[idx]:
                                    tempControl.append(0)
                                elif element > desiredKey[idx]:
                                    tempControl.append(-1)
                                    control_mag[sampleidx] += 1
                                elif element < desiredKey[idx]:
                                    tempControl.append(1)
                                    control_mag[sampleidx] += 1
                cellidx += (self.currRadius*2)+1

            # finish the row of control if not done
            while (len(tempControl) < self._embed_dim):
                tempControl.append(0)

            control.append([tempControl])

        # for all other states of traj, choose no control
        for sampleidx in range(len(x)):
            for _ in range(1, self._timesteps):
                control[sampleidx].append(np.zeros(self._embed_dim))

        return np.array(control)

    def _evolve(self, x):
        evolved = []

        for sample in x:
            sampleResult = []
            for cellidx in range(0, self.ROW_LENGTH):
                neighborhood = ""

                # to the left
                for neg_x in range(self.currRadius, 0, -1):
                    neighborhood += str(sample[cellidx - neg_x])

                # directly above
                neighborhood += str(sample[cellidx])

                # to the right
                for pos_x in range(1, self.currRadius+1):
                    pos = (cellidx + pos_x) % self.ROW_LENGTH
                    neighborhood += str(sample[pos])

                # if seen before, use knowledge
                if (self.radiiTables[self.currRadius][neighborhood]) != None:
                    sampleResult.append(
                        self.radiiTables[self.currRadius][neighborhood])

                # else -> random guess between 0/1
                else:
                    sampleResult.append(np.random.randint(0, 1))
            evolved.append(sampleResult)
        return evolved

    def predict(self, x0, timesteps, **kwargs):
        preds = [x0]

        for _ in range(timesteps-1):
            preds.append(self._evolve(preds[-1]))

        preds = np.array(preds)
        preds = np.transpose(preds, (1, 0, 2))

        return preds



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import LineCollection
from mpl_toolkits.mplot3d.art3d import Line3DCollection
from sklearn.decomposition import PCA

MAX_LINES = 30

def _plot2d(trajs_grid, labels: list = None):
    fig = plt.figure()
    ax = fig.add_subplot()
    for i, trajs in enumerate(trajs_grid):
        if labels is None:
            line_collection = LineCollection(trajs, color=f"C{i}")
        else:
            line_collection = LineCollection(trajs, color=f"C{i}", label=labels[i])
        ax.add_collection(line_collection)
        ax.scatter(trajs[:, 0, 0], trajs[:, 0, 1], color=f"C{i}", marker="o")
    minima = trajs_grid.min(axis=(0, 1, 2))
    maxima = trajs_grid.max(axis=(0, 1, 2))
    ax.set_xlim(minima[0], maxima[0])
    ax.set_ylim(minima[1], maxima[1])
    if labels is not None:
        ax.legend()
    return fig, ax


def _plot3d(grid, labels: list = None):
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    for i, trajs in enumerate(grid):
        if labels is None:
            line_collection = Line3DCollection(trajs, color=f"C{i}")
        else:
            line_collection = Line3DCollection(trajs, color=f"C{i}", label=labels[i])
        ax.add_collection(line_collection)
        ax.scatter(trajs[:, 0, 0], trajs[:, 0, 1], trajs[:, 0, 2], color=f"C{i}", marker="o")
    minima = grid.min(axis=(0, 1, 2))
    maxima = grid.max(axis=(0, 1, 2))
    ax.set_xlim(minima[0], maxima[0])
    ax.set_ylim(minima[1], maxima[1])
    ax.set_zlim(minima[2], maxima[2])
    if labels is not None:
        ax.legend()
    return fig, ax


def _plot4d(trajs_grid, pca: PCA, labels: list = None):
    assert pca.n_components == 2 or pca.n_components == 3
    trajs_grid = _apply_pca_to_grid(trajs_grid, pca)
    if pca.n_components == 2:
        return _plot2d(trajs_grid, labels)
    else:
        return _plot3d(trajs_grid, labels)


def make_pca(trajs: np.ndarray, n_components=3):
    dim = trajs.shape[-1]
    X = trajs.reshape((-1, dim))
    pca = PCA(n_components=n_components)
    pca.fit(X)
    return pca


def _apply_pca_to_grid(trajs_grid, pca) -> np.ndarray:
    dim = trajs_grid.shape[-1]
    X = trajs_grid.reshape((-1, dim))
    result = pca.transform(X)
    target_shape = list(trajs_grid.shape)
    target_shape[-1] = pca.n_components
    result = result.reshape(target_shape)
    return result


def plot(grid: list[np.ndarray], target_dim: int = 3, max_lines=MAX_LINES, labels: list[str] = None):
    grid = np.array([x[:max_lines] for x in grid])
    dim = grid.shape[-1]
    assert target_dim <= dim
    assert target_dim <= 3
    if dim == 2:
        return _plot2d(grid, labels)
    elif dim == 3 and target_dim == 3:
        return _plot3d(grid, labels)
    else:
        pca = make_pca(grid, n_components=target_dim)
        return _plot4d(grid, pca, labels)



from setuptools import setup, find_packages


setup(
    name="dynadojo",
    version="0.1.0",
    description="An extensible platform for benchmarking sample efficiency in dynamical system identification",
    url="https://github.com/FlyingWorkshop/DynaDojo",
    author="Logan Mondal Bhamidipaty",
    author_email="loganmb@stanford.edu",
    license="MIT",
    python_requires=">=3.10",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10"
    ],
    install_requires=[
        "tqdm>=4.65.0",
        "numpy>=1.23.5",
        "torch>=2.0.1",
        "torchdiffeq>=0.2.3",
        "joblib>=1.2.0",
        "keras>=2.12.0",
        "tensorflow",
        "seaborn>=0.12.2",
        "scipy>=1.10.1",
        "scikit-learn>=1.2.2",
        "cellpylib",
        "matplotlib",
        "pandas",
        "networkx",
    ],
    packages=find_packages(),
)


import warnings
import pandas as pd
import numpy as np

from .utils.plotting import plot_target_loss, plot_metric

from joblib import Parallel, delayed

from .abstractions import Task, AbstractSystem, AbstractModel


class FixedError(Task):
    def __init__(self, L: list[int], t: int, max_control_cost_per_dim: int, control_horizons: int,
                 system_cls: type[AbstractSystem], reps: int, test_examples: int, test_timesteps: int, target_loss: float,
                 system_kwargs: dict = None,
                 max_samples=10000, E: int | list[int] = None):

        if isinstance(E, list):
            assert (len(L) == len(E))
            for idx, l in enumerate(L):
                assert (E[idx] >= l)
        elif isinstance(E, int):
            assert (E >= np.max(L))

        # the dimensions have to be sorted, we need a guarantee that they are sorted
        self._target_loss = target_loss
        self.max_samples = max_samples

        self.X = None
        self.rep_id = None
        self.result = None
        self.samples_needed = {}

        super().__init__([1], L, E, t, max_control_cost_per_dim, control_horizons, system_cls, reps, test_examples, test_timesteps, system_kwargs=system_kwargs)

    def _evaluate_n(self, system, model_cls, test, n, max_control_cost, noisy, fit_kwargs, act_kwargs, model_kwargs):
        x = None

        df = pd.DataFrame(self.result)
        df = df.loc[(df['rep'] == self.rep_id) & (df['n'] == n) & df['latent_dim'] == system.latent_dim]
        if len(df) == 1:
            return df.iloc[0]['loss']

        if self.X is None:
            self.X = self._gen_trainset(system, n, self._T, noisy)
            x = self.X

        # if already generated n samples, pull from existing training set; else add more samples
        else:
            if (len(self.X) >= n):
                x = self.X[:int(n), : , :]
            else:
                to_add = self._gen_trainset(
                    system, n - len(self.X), self._T, noisy)
                self.X = np.concatenate((self.X, to_add), axis=0)
                x = self.X

        model = model_cls(system.embed_dim, self._T, max_control_cost, **model_kwargs)
        total_cost = self._fit_model(system, model, x, self._T,
                                     max_control_cost, fit_kwargs, act_kwargs, noisy)


        pred = model.predict_wrapper(test[:, 0], self._test_timesteps)

        loss = system.calc_error_wrapper(pred, test)

        self._append_result(self.result, self.rep_id, n, system.latent_dim, system.embed_dim, self._T, loss, total_cost)

        return loss

    def _gen_bounds(self, system, model_cls, test, l, max_control_cost, noisy, fit_kwargs, act_kwargs, model_kwargs):
        lower = 1
        upper = 3
        predicted = 2

        if self.samples_needed:
            prev_largest_dim = np.max(list(self.samples_needed.keys()))
            lower = self.samples_needed[prev_largest_dim]

            # adaptively predict based on median of samples needed in lower dims
            if len(self.samples_needed.keys()) > 1:
                # p = np.polyfit(list(self.samples_needed.keys()), list(self.samples_needed.values()), deg=len(list(self.samples_needed.keys())) - 1)
                p = self.samples_needed[prev_largest_dim] * 2
                predict_func = np.poly1d(p)
                predicted = int(round(predict_func(l)))

            # with only one data dim seens, assume a very small slope
            else:
                predicted = self.samples_needed[prev_largest_dim] + 1

            if predicted < lower:
                predicted = lower+1
            upper = (2 * predicted) - lower

        # if lower meets target loss, decrease to be lower
        lower_error = self._evaluate_n(system, model_cls, test, lower, max_control_cost, noisy, fit_kwargs, act_kwargs, model_kwargs)
        while lower_error <= self._target_loss:
            if lower == 1:
                break
            lower = lower // 2
            lower_error = self._evaluate_n(system, model_cls, test, lower, max_control_cost, noisy, fit_kwargs, act_kwargs, model_kwargs)

        # increase upper until it satisfies target loss or exceeds max_samples
        while upper <= self.max_samples:
            if upper == self.max_samples:
                max_samples_error = self._evaluate_n(
                    system, model_cls, test, self.max_samples, max_control_cost, noisy, fit_kwargs, act_kwargs, model_kwargs)
                if max_samples_error <= self._target_loss:
                    break
                else:
                    warnings.warn(f'Rep #{self.rep_id} for l={system.latent_dim} and e={system.embed_dim} failed: Fixed error not achieved within max samples. At {self.max_samples} samples, error is still {max_samples_error}, which is above {self._target_loss}')
                    return (lower, None) 

            else:
                upper_error = self._evaluate_n(
                    system, model_cls, test, upper, max_control_cost, noisy, fit_kwargs, act_kwargs, model_kwargs)
                
                if upper_error <= self._target_loss:
                    break
                else:
                    if upper == lower:
                        temp_upper = lower+1
                    else:
                        temp_upper = round((2 * (upper - lower)) + upper)
                    lower = upper
                    upper = min(self.max_samples, temp_upper)
        

        return (lower, upper)

    def _search(self, system, model_cls, test, lower, upper, max_control_cost, noisy, fit_kwargs, act_kwargs, model_kwargs):
        lowest_needed = None
       
        while (lower != upper):
            mid = int((lower + upper) / 2)

            mid_error = self._evaluate_n(
                system, model_cls, test, mid, max_control_cost, noisy, fit_kwargs, act_kwargs, model_kwargs)


            if mid_error > self._target_loss:
                if lower == mid:
                    lower = mid + 1
                else:
                    lower = mid
            else:
                upper = mid
                lowest_needed = mid
        
        return lowest_needed
               

    def evaluate(self,
                 model_cls: type[AbstractModel],
                 model_kwargs: dict = None,
                 fit_kwargs: dict = None,
                 act_kwargs: dict = None,
                 in_dist=True,
                 noisy=False,
                 id=None,
                 ):

        model_kwargs = model_kwargs or {}
        fit_kwargs = fit_kwargs or {}
        act_kwargs = act_kwargs or {}

        data = []

        def _do_rep(rep_id, latent_dim, embed_dim):

            self.rep_id = rep_id

            self.result = {k: [] for k in ["rep", "n", "latent_dim", "embed_dim", "timesteps", "control_horizons", "loss", "total_cost"]}
            print(f"{latent_dim=}, {embed_dim=}, timesteps={self._T}, control_horizons={self._control_horizons}, {rep_id=}, {id=}")

            system = None
            system = self._set_system(system, latent_dim, embed_dim)
            max_control_cost = self._max_control_cost_per_dim * latent_dim
            
            test = self._gen_testset(system, in_dist)

            lower, upper = self._gen_bounds(system, model_cls, test, latent_dim, max_control_cost, noisy, fit_kwargs, act_kwargs, model_kwargs)
            if upper is not None:
                self._search(system, model_cls, test, lower, upper, max_control_cost, noisy, fit_kwargs, act_kwargs, model_kwargs)

            self.X = None 
            return pd.DataFrame(self.result)

        for idx, latent_dim in enumerate(self._L):
            self.X = None
            if not self._E:
                embed_dim = latent_dim
            elif isinstance(self._E, int):
                embed_dim = self._E
            else:
                embed_dim = self._E[idx]

            temp_df = Parallel(n_jobs=1)(delayed(_do_rep)(rep_id, latent_dim, embed_dim) for rep_id in range(self._reps))
            temp_df = pd.concat(temp_df, ignore_index=True)

            self.samples_needed[latent_dim] = int(round(np.median(temp_df["n"])))

            data.append(temp_df)
       
        self.samples_needed = None
        self.X = None
        self.result = None

        data = pd.concat(data,  ignore_index=True)
        data["id"] = id or next(self._id)

        return data

    def plot(self, data):
        plot_target_loss(data, "latent_dim", "n", target_loss=self._target_loss, ylabel=r'$n$')


class FixedComplexity(Task):
    def __init__(self, N: list[int], l: int, e: int, t: int, max_control_cost_per_dim: int, control_horizons: int,
                 system_cls: type[AbstractSystem], reps: int, test_examples: int, test_timesteps: int, system_kwargs: dict = None):
        L = [l]
        E = e
        T = [t]
        super().__init__(N, L, E, T, max_control_cost_per_dim, control_horizons,
                         system_cls, reps, test_examples, test_timesteps, system_kwargs=system_kwargs)

    @staticmethod
    def plot(data):
        plot_metric(data, "n", "loss", xlabel=r'$n$')

class FixedTrainSize(Task):
    def __init__(self, n: int, L: list[int], E: list[int] | int | None, T: list[int], max_control_cost_per_dim: int, control_horizons: int,
                 system_cls: type[AbstractSystem], reps: int, test_examples: int, test_timesteps: int, system_kwargs: dict = None):
        N = [n]
        super().__init__(N, L, E, T, max_control_cost_per_dim, control_horizons,
                         system_cls, reps, test_examples, test_timesteps, system_kwargs=system_kwargs)

    @staticmethod
    def plot(data: pd.DataFrame):
        plot_metric(data, "latent_dim", "loss", xlabel=r'$L$', log=False)



import numpy as np
from ..abstractions import AbstractModel

from sklearn.linear_model import LinearRegression as SKLinearRegression


class LinearRegression(AbstractModel):
    def __init__(self, embed_dim, timesteps, max_control_cost):
        super().__init__(embed_dim, timesteps, max_control_cost)
        self.A_hat = []
        self.U = None
        self.model = None

    def fit(self, x: np.ndarray, **kwargs):

        N, T, D = x.shape
        X_train = x[:, :-1, :].reshape(N * (T - 1), D)
        y_train = x[:, 1:, :].reshape(N * (T - 1), D)

        self.model = SKLinearRegression()

        if self.U is not None:
            U = np.reshape(U, (U.shape[0] * U.shape[1], U.shape[2])).T

            X_train = np.vstack((X_train, U))

        self.model.fit(X_train, y_train)
        self.A_hat = self.model.coef_

    def act(self, x, **kwargs):
        self.U = np.random.uniform(-1, 1,
                                   [len(x[0]), self._timesteps, self._embed_dim])
        self.U = np.array(self.U)

        self.U = (self.U / np.linalg.norm(self.U, axis=-1)
                  [:, :, np.newaxis]) * self._max_control_cost

        return self.U

    def predict(self, x0, timesteps, **kwargs):
        preds = []
        traj = [x0.T]
        for _ in range(timesteps-1):
            traj.append(self.A_hat @ traj[-1])
        preds.append(traj)
        preds = np.squeeze(np.array(preds), 0)
        preds = np.transpose(preds, (2, 0, 1))
        return preds


class ManualLinearRegression(AbstractModel):
    def __init__(self, embed_dim, timesteps, control_constraint):
        super().__init__(embed_dim, timesteps, control_constraint)
        self.A_hat = []
        self.U = None
        self.model = None

    def fit(self, x: np.ndarray, **kwargs):
        X = x[:, :-1, :]
        Y = x[:, 1:, :]

        X = np.reshape(X, (X.shape[0] * (X.shape[1]), X.shape[2])).T
        Y = np.reshape(Y, (Y.shape[0] * (Y.shape[1]), Y.shape[2])).T

        # NOTE: pseudo-inverse does not work with low rank matrix
        if self.U is not None:
            U = np.reshape(U, (U.shape[0] * U.shape[1], U.shape[2])).T

            d = X.shape[-1]
            matrix = Y@(np.linalg.pinv(np.vstack((X, U))))
            self.A_hat = matrix[:, :d]

        else:
            pinv = np.linalg.pinv(X)
            self.A_hat = Y@pinv

    def act(self, x, **kwargs):
        self.U = np.random.uniform(-1, 1,
                                   [len(x[0]), self._timesteps, self._embed_dim])
        self.U = np.array(self.U)

        self.U = (self.U / np.linalg.norm(self.U, axis=-1)
                  [:, :, np.newaxis]) * self._max_control_cost

        return self.U

    def predict(self, x0, timesteps, **kwargs):
        preds = []
        traj = [x0.T]
        for _ in range(timesteps-1):
            traj.append(self.A_hat @ traj[-1])
        preds.append(traj)
        preds = np.squeeze(np.array(preds), 0)
        preds = np.transpose(preds, (2, 0, 1))
        return preds



import numpy as np
from joblib import Parallel, delayed
from scipy.integrate import solve_ivp
from scipy.interpolate import interp1d
from scipy.optimize import minimize, NonlinearConstraint

from .lds import LDSSystem

RNG = np.random.default_rng()


class SNNSystem(LDSSystem):

    def __init__(self, latent_dim, embed_dim,
                 init_cond_in_dist_range=(0, 1),
                 init_cond_out_dist_range=(-1, 0)
                 ):
        assert embed_dim >= 2 * latent_dim, "REQUIRED: embed_dim ≥ 2 * latent_dim"
        super().__init__(latent_dim, embed_dim,
                         init_cond_in_dist_range=init_cond_in_dist_range,
                         init_cond_out_dist_range=init_cond_out_dist_range,
                         )
        c = lambda t: np.zeros(latent_dim)
        self.LDS = LinearDynamicalSystem(self.A, self.B, c)
        self.SNN = SpikingNeuralNetwork(self.LDS,
                                        N=embed_dim,
                                        max_error=1 / 2)  # hyperparam: max_error

    def make_init_conds(self, n: int, in_dist=True) -> np.ndarray:
        return super().make_init_conds(n, in_dist)

    def make_data(self, init_conds: np.ndarray,
                  control: np.ndarray,
                  timesteps: int,
                  noisy=False) -> np.ndarray:
        time = np.linspace(0, 1, num=timesteps)  # hyperparam: [t0, tf]

        def get_trajectory(x0, u):
            u_interp = interp1d(time, u.T, kind=3)  # TODO: only works if u[-1] = u(t_f)
            self.SNN.set_control(u_interp)
            results = self.SNN.simulate(x0, time)
            t_snn = np.hstack(results["t"])
            x_snn = self.SNN.Phi @ (np.hstack(results["rv"])[:self.SNN.N, :])
            v_snn = np.hstack(results["rv"])[self.SNN.N:, :]
            return v_snn

        data = Parallel(n_jobs=4)(delayed(get_trajectory)(x0, u) for x0, u in zip(init_conds, control))
        # PARALLELIZE: This is slower than serial for some reason?
        # https://stackoverflow.com/questions/9786102/how-do-i-parallelize-a-simple-python-loop
        # https://colab.research.google.com/github/QuantEcon/lecture-python-programming.notebooks/blob/master/parallelization.ipynb

        # trajs = []
        # for x0, u in tqdm(zip(init_conds, control), total=len(init_conds)):
        #     trajs.append(get_trajectory(x0, u))

        data = np.array(data)
        data = np.transpose(data, axes=(0, 2, 1))
        return data


class LinearDynamicalSystem:

    def __init__(self, A, B, control):
        # dx/dt = Ax + Bc
        self.dim = A.shape[0]
        self.A = A
        self.B = B
        self.c = control
        self.x_dot = self._construct_LDS()

    def _construct_LDS(self):
        return lambda t, x: self.A @ x + self.B @ self.c(t)

    def simulate(self, x_0, t_eval):
        t_0, t_f = t_eval[[0, -1]]
        max_step = (t_eval[1] - t_eval[0]) / 10
        results = solve_ivp(self.x_dot,
                            [t_0, t_f],
                            x_0,
                            method=['RK45', 'RK23', 'DOP853', 'Radau', 'BDF', 'LSODA'][0],
                            t_eval=t_eval,
                            max_step=max_step,
                            # rtol = 1e-3,
                            # atol= 1e-6,
                            vectorized=False
                            )
        return results


class SpikingNeuralNetwork:

    def __init__(self, LDS, N=None, max_error=1 / np.sqrt(2)):
        self.LDS = LDS
        self.N = N if N else 2 * LDS.dim
        self.threshold = (2 * max_error) ** 2 / 2.0
        self.Phi = self._generate_decoders()
        self.control = lambda t: np.zeros(self.N)
        self.rv_dot = self._construct_SNN()

    def _generate_decoders(self):
        d = self.LDS.dim
        magnitude = np.sqrt(2 * self.threshold)  # |d| = 2*max_error , T = |d|^2 / 2

        def min_distance(x_flat):
            # objective: maximize the minimum pairwise distance
            from sklearn.metrics import pairwise_distances as pdist
            X = np.reshape(x_flat, (d, self.N))
            dist = pdist(X.T)
            return -np.min(dist[np.tril_indices_from(dist, k=-1)])

        def threshold(x_flat):
            X = np.reshape(x_flat, (d, self.N))
            return np.linalg.norm(X, axis=0)

        def generate_decoder(magnitude):
            X = np.random.normal(size=(d, self.N))
            X *= magnitude / np.linalg.norm(X, axis=0)
            result = minimize(min_distance,
                              X.flatten(),
                              constraints=NonlinearConstraint(threshold,
                                                              magnitude,
                                                              magnitude),
                              tol=1e-10,
                              )
            return result.x.reshape(d, self.N)

        D = generate_decoder(magnitude)
        while ~np.all((np.max(D, axis=1) > 0) & (np.min(D, axis=1) < 0)):
            D = generate_decoder(magnitude)

        return D

    def _construct_SNN(self):
        d = self.LDS.dim
        VV = self.Phi.T @ self.LDS.A @ np.linalg.pinv(self.Phi.T)
        VR = self.Phi.T @ (self.LDS.A + np.eye(d)) @ self.Phi
        VD = -self.Phi.T @ self.Phi
        VC = self.Phi.T @ self.LDS.B
        RR = -np.eye(self.N)
        RD = np.eye(self.N)
        RVRV = np.vstack([np.hstack([RR, np.zeros((self.N, self.N))]),
                          np.hstack([VR, VV])])
        RVD = np.vstack([RD, VD])
        self.RVD = RVD
        RVC = np.vstack([np.zeros(VC.shape), VC])
        return lambda t, rv: RVRV @ rv + np.hstack([np.zeros(self.N), self.control(t)])  # RVC@self.LDS.c(t)

    def set_control(self, control):
        self.control = control

    def simulate(self, x_0, t_eval):

        def event_list_generator():
            event_list = []
            for i in range(self.N):
                e = (lambda i: lambda t, rv: rv[self.N + i] - self.threshold)(i)
                e.terminal = True
                e.direction = 1
                event_list += [e]
            return event_list

        t_0, t_f = t_eval[[0, -1]]
        max_step = (t_eval[1] - t_eval[0]) / 10

        rv_0 = np.hstack([x_0, np.zeros(self.N)])
        events = event_list_generator()

        all_results = {
            "t": [],
            "rv": [],
            'spikes': []
        }

        while t_0 < t_f:

            results = solve_ivp(self.rv_dot,
                                [t_0, t_f],
                                rv_0,
                                method=['RK45', 'RK23', 'DOP853', 'Radau', 'BDF', 'LSODA'][0],
                                events=events,
                                t_eval=t_eval[t_0 <= t_eval],
                                max_step=max_step,
                                # rtol = 1e-3,
                                # atol= 1e-6,
                                vectorized=False)

            if np.any(results.t_events):
                t_spike = np.hstack(results.t_events)[0]
                n_idx = np.nonzero(np.array(results.t_events, dtype=object))[0][0]
                # print('Neuron {}: {} s'.format(n_idx, t_spike))
                all_results['spikes'].append((n_idx, t_spike))

                # spike update
                t_0 = t_spike
                delta = np.zeros(self.N)
                delta[n_idx] = 1
                rv_0 = results.y_events[n_idx][0] + self.RVD @ delta

            if np.any(results.t):
                all_results["t"] += [results.t]
                all_results["rv"] += [results.y]
                if results.t[-1] == t_f:
                    break

        return all_results
